/**
 * Created by janis dähne.
 */
import {ActionBase, ActionType} from "../../reducers/tutorViewSite/tutorViewSiteActionTypes";
import {
  RESET_userSpecificDataAction,
  ResetAction,
  SET_botLeftPanelViewSizeAction,
  SET_botRightPanelViewSizeAction,
  SET_bottomLeftSelectedTabIndexAction,
  SET_bottomRightSelectedTabIndexAction, SET_codeTemplate_templateFilesAction,
  SET_comesFromOwnExercisesSiteAction,
  SET_dummyAction, SET_feedbackForStudentPanelSelectedSubTabIndexAction,
  SET_hDividerPosOffsetAction,
  SET_isCodeEditorSettingsDialogDisplayedAction, SET_isJustRunProgramArgsInputDisplayedAction,
  SET_isLoadingAction,
  SET_isViewTestDialogDisplayedAction, SET_justRunProgramArgsAction,
  SET_selectedConsoleViewTestAction,
  SET_topLeftPanelViewSizeAction,
  SET_topLeftSelectedTabIndexAction,
  SET_topRightPanelViewSizeAction,
  SET_topRightSelectedTabIndexAction,
  SET_vDividerPosOffsetAction
} from "../../reducers/tutorViewSite/tutorViewSiteReducer";
import {PanelViewSize} from "../../../components/panelSystem/panelGrid/panelGrid";
import {AwaitActions, MultiActions} from "../types";
import {
  getAllAfterSubmitTestResultsAsTutor,
  getAllUserAfterTestResultsAsTutorAsnc,
  getAllUserTestResultsAsTutorAsnc,
  getExerciseTutorViewAsync,
  getSingleAssessmentTutorViewAsync,
  getUserAfterSolutionAsTutorViewAsync,
  getUserSolutionAsTutorViewAsync,
  getUserSubmissionTestResultsAsTutorAsync
} from "./tutorViewSiteCrudActions";
import {createAndRefreshAssetUrlsAndObjects, freeAssetUrlsAndObjects} from "../../../helpers/assetHelper";
import {getTestTypesAsync} from "../testTypesActions";
import {AssessedSubmitTestFrontendExclusive} from "../../../types/testResults";
import {RunnableCustomTestFrontendExclusive} from "../../../types/customTests";
import {ExerciseCodeTemplateFileFullBase, RunnableExerciseTestFrontendExclusive} from "../../../types/exerciseEditor";
import {SolutionDoExerciseFullBase, SolutionFileDoExerciseFullBase} from "../../../types/exerciseSolution";
import {SET_userSolution_solutionFilesAction} from "../../reducers/tutorViewSite/subSets/setUserSolutionReducer";
import {setSelectedTabIdUserSolutionEditorTutorView} from "./editorActions/userSolutionEditorActions";
import {
  SET_tutorEditorSolution_mainFileIdAction,
  SET_tutorEditorSolution_solutionFilesAction
} from "../../reducers/tutorViewSite/subSets/setTutorEditorSolutionReducer";
import {
  setHasFileChangedMapTutorEditor,
  setHasUnsavedChangesTutorEditor,
  setSelectedTabIdTutorEditor
} from "./editorActions/tutorEditorActions";
import Logger from "../../../helpers/logger";
import {setSubmitTestsForTutor} from "./sub/submitTestActions";
import {mergeNormalTestResults, mergeSubmitTestResults} from "../../../helpers/convertersAndTransformers";
import {setNormalTestsForTutor} from "./sub/normalTestActions";
import {
  AceEditorsStorageHelper,
  AceStoredTabLike
} from '../../../components/codeEditors/aceEditors/aceEditorsStorageHelper'
import {
  tutorViewCodeTemplateEditorPanelViewEditorId,
  tutorViewEditorPanelViewEditorId,
  tutorViewUserAfterSolutionEditorPanelViewEditorId,
  tutorViewUserSolutionEditorPanelViewEditorId
} from '../../../constants'
import {SET_userAfterSolution_solutionFilesAction} from '../../reducers/tutorViewSite/subSets/setUserAfterSolutionReducer'
import {setSelectedTabIdAfterEditorTutorView} from './editorActions/userAfterSolutionEditorTutorActions'
import {AfterSolutionDoExerciseFullBase} from '../../../types/exerciseAfterSolution'
import {
  getGlobalSystemSettingsAsync,
  getTestServerPublicSettingsAsync
} from '../globalSystemSettings/globalSystemSettingsCrudActions'
import {setSelectedTabIdCodeTemplateEditorTutorView} from './editorActions/codeTemplateEditorTutorActions'
import {getPLangsAsync} from '../plangSettingsView/pLangsCrudActions'


export function setSelectedConsoleViewTest(selectedConsoleViewTest: RunnableExerciseTestFrontendExclusive
  | RunnableCustomTestFrontendExclusive | AssessedSubmitTestFrontendExclusive | null): SET_selectedConsoleViewTestAction {
  return {
    type: ActionType.SET_selectedConsoleViewTest,
    selectedConsoleViewTest
  }
}

//used to open/close the user solution editor tabs
export function setSolutionFilesUserSolutionTutorView(solutionFiles: ReadonlyArray<SolutionFileDoExerciseFullBase>): SET_userSolution_solutionFilesAction {
  return {
    type: ActionType.SET_userSolution_solutionFiles,
    solutionFiles
  }
}

export function setSolutionFilesUserAfterSolutionTutorView(afterSolutionFiles: ReadonlyArray<SolutionFileDoExerciseFullBase>): SET_userAfterSolution_solutionFilesAction {
  return {
    type: ActionType.SET_userAfterSolution_solutionFiles,
    afterSolutionFiles
  }
}

export function setSolutionFilesCodeTemplateTutorView(codeTemplateId: number, templateFiles: ReadonlyArray<ExerciseCodeTemplateFileFullBase>): SET_codeTemplate_templateFilesAction {
  return {
    type: ActionType.SET_codeTemplate_templateFiles,
    templateFiles,
    codeTemplateId
  }
}

export function setSolutionFilesTutorEditor(solutionFiles: ReadonlyArray<SolutionFileDoExerciseFullBase>): SET_tutorEditorSolution_solutionFilesAction {
  return {
    type: ActionType.SET_tutorEditorSolution_solutionFiles,
    solutionFiles
  }
}

//used for resetting the tutor editor
export function setMainFileIdTutorEditor(mainFileId: number): SET_tutorEditorSolution_mainFileIdAction {
  return {
    type: ActionType.SET_tutorEditorSolution_mainFileId,
    mainFileId
  }
}

export function resetTutorEditorToUserSolution(userSolution: SolutionDoExerciseFullBase | AfterSolutionDoExerciseFullBase): AwaitActions {
  return async (dispatch, getState) => {

    const newSolutionFiles = [...userSolution.solutionFiles] as ReadonlyArray<SolutionFileDoExerciseFullBase>
    dispatch(setSolutionFilesTutorEditor(newSolutionFiles))

    const pLang = getState().pLangsState.pLangs.find(p => p.id === userSolution.pLangId)

    if (!pLang) {
      Logger.error('could not find plang')
    }

    //set the editor storage
    //add a dummy because we have no access to the real editor tabs... but we only need the ids
    // noinspection TsLint
    const storedTabs = newSolutionFiles.map<AceStoredTabLike>(p => {
      return {
        id: p.id,
        content: p.content,
        language: pLang.internalName
      }
    })
    AceEditorsStorageHelper.removeEditor(tutorViewEditorPanelViewEditorId)
    AceEditorsStorageHelper.addEditorWithTabs(tutorViewEditorPanelViewEditorId, storedTabs)

    //we setting main file or setHasUnsavedChangesTutorEditor or setHasFileChangedMapTutorEditor
    //will force the editor to rerender and update the session
    //else we get the right content but the undo stack is wrong (because we have still the old session)
    dispatch(setMainFileIdTutorEditor(userSolution.mainFileId))


    //also reset the editor changes state...
    dispatch(setHasUnsavedChangesTutorEditor(false))
    dispatch(setHasFileChangedMapTutorEditor({}))

    //clear all new results
    dispatch(resetTutorNormalTestResults())
    dispatch(resetTutorSubmitTestResults())

    //select the first file if any

    if (newSolutionFiles.length > 0) {
      dispatch(setSelectedTabIdTutorEditor(storedTabs[0].id))
    }

  }
}

export function resetTutorNormalTestResults(): MultiActions {
  return (dispatch, getState) => {
    dispatch(
      setNormalTestsForTutor(mergeNormalTestResults(getState().tutorViewSiteState.normalTestsForTutor, [], true)))
    dispatch(setSelectedConsoleViewTest(null))
  }
}

export function resetTutorSubmitTestResults(): MultiActions {
  return (dispatch, getState) => {
    dispatch(
      setSubmitTestsForTutor(mergeSubmitTestResults(getState().tutorViewSiteState.submitTestsForTutor, [], true)))
    dispatch(setSelectedConsoleViewTest(null))
  }
}


export function setTutorViewJustRunProgramArgs(justRunProgramArgs: string): SET_justRunProgramArgsAction {
  return {
    type: ActionType.SET_justRunProgramArgs,
    justRunProgramArgs
  }
}

export function setTutorViewIsJustRunProgramArgsInputDisplayed(isJustRunProgramArgsInputDisplayed: boolean): SET_isJustRunProgramArgsInputDisplayedAction {
  return {
    type: ActionType.SET_isJustRunProgramArgsInputDisplayed,
    isJustRunProgramArgsInputDisplayed
  }
}

//--- dialogs

export function setTutorViewIsCodeEditorSettingsDialogDisplayed(isDisplayed: boolean): SET_isCodeEditorSettingsDialogDisplayedAction {
  return {
    type: ActionType.SET_isCodeEditorSettingsDialogDisplayed,
    isDisplayed
  }
}

export function setIsViewTestDialogDisplayed(isViewTestDialogDisplayed: boolean): SET_isViewTestDialogDisplayedAction {
  return {
    type: ActionType.SET_isViewTestDialogDisplayed,
    isViewTestDialogDisplayed
  }
}

//--- grid system, tab indices
export function setTopRightSelectedTabIndex(topRightSelectedTabIndex: number): SET_topRightSelectedTabIndexAction {
  return {
    type: ActionType.SET_topRightSelectedTabIndex,
    topRightSelectedTabIndex
  }
}

export function setTopLeftSelectedTabIndex(topLeftSelectedTabIndex: number): SET_topLeftSelectedTabIndexAction {
  return {
    type: ActionType.SET_topLeftSelectedTabIndex,
    topLeftSelectedTabIndex
  }
}

export function setBottomRightSelectedTabIndex(bottomRightSelectedTabIndex: number): SET_bottomRightSelectedTabIndexAction {
  return {
    type: ActionType.SET_bottomRightSelectedTabIndex,
    bottomRightSelectedTabIndex
  }
}

export function setBottomLeftSelectedTabIndex(bottomLeftSelectedTabIndex: number): SET_bottomLeftSelectedTabIndexAction {
  return {
    type: ActionType.SET_bottomLeftSelectedTabIndex,
    bottomLeftSelectedTabIndex
  }
}

export function setFeedbackForStudentPanelSelectedSubTabIndex(feedbackForStudentPanelSelectedSubTabIndex: number): SET_feedbackForStudentPanelSelectedSubTabIndexAction {
  return {
    type: ActionType.SET_feedbackForStudentPanelSelectedSubTabIndex,
    feedbackForStudentPanelSelectedSubTabIndex
  }
}


//--- gripper
export function setDoExerciseHDividerPosOffset(hDividerPosOffset: number): SET_hDividerPosOffsetAction {
  return {
    type: ActionType.SET_hDividerPosOffset,
    hDividerPosOffset
  }
}

export function setDoExerciseVDividerPosOffset(vDividerPosOffset: number): SET_vDividerPosOffsetAction {
  return {
    type: ActionType.SET_vDividerPosOffset,
    vDividerPosOffset
  }
}

//--- panel view sizes

export function setTopRightPanelViewSize(topRightPanelViewSize: PanelViewSize): SET_topRightPanelViewSizeAction {
  return {
    type: ActionType.SET_topRightPanelViewSize,
    topRightPanelViewSize
  }
}

export function setTopLeftPanelViewSize(topLeftPanelViewSize: PanelViewSize): SET_topLeftPanelViewSizeAction {
  return {
    type: ActionType.SET_topLeftPanelViewSize,
    topLeftPanelViewSize
  }
}

export function setBotRightPanelViewSize(botRightPanelViewSize: PanelViewSize): SET_botRightPanelViewSizeAction {
  return {
    type: ActionType.SET_botRightPanelViewSize,
    botRightPanelViewSize
  }
}

export function setBotLeftPanelViewSize(botLeftPanelViewSize: PanelViewSize): SET_botLeftPanelViewSizeAction {
  return {
    type: ActionType.SET_botLeftPanelViewSize,
    botLeftPanelViewSize
  }
}


export function loadTutorViewSite(releaseId: number, userId: number, pLangId: number,
                                  comesFromOwnExercisesSite: boolean
): AwaitActions {
  return async (dispatch, getState) => {

    dispatch(resetTutorViewSite())

    dispatch(setComesFromOwnExercisesSite(comesFromOwnExercisesSite))

    //load exercise
    //with tests, custom tests
    await dispatch(getExerciseTutorViewAsync(releaseId, userId, pLangId))

    //load test types, system settings ... same as do exercise site loading
    await dispatch(getGlobalSystemSettingsAsync())

    await dispatch(getTestServerPublicSettingsAsync())

    await dispatch(getTestTypesAsync())

    await dispatch(getPLangsAsync())

    await dispatch(_loadTutorViewData(releaseId, userId, pLangId))
  }
}

function _loadTutorViewData(releaseId: number, userId: number, pLangId: number): MultiActions {
  return async (dispatch, getState) => {

    //get solution
    //this sets the user solution
    //and also the tutor editor solution (which is a copy of the user solution initially)
    await dispatch(getUserSolutionAsTutorViewAsync(releaseId, userId, pLangId))

    //custom, test results
    await dispatch(getAllUserTestResultsAsTutorAsnc(releaseId, userId, pLangId))

    //submit test results only if they can be present
    if (!getState().tutorViewSiteState.exercise.exerciseRelease.isReleased
      && getState().tutorViewSiteState.exercise.exerciseRelease.hasAutomaticAssessmentFinished) {
      try {
        //this might fail if the assessment has not started yet...
        await dispatch(getUserSubmissionTestResultsAsTutorAsync(releaseId, userId, pLangId))
      } catch (err) {
        //could be ok
        Logger.warn(
          "could not get submission test result as tutor, error (automatic assessment might not be finished yet): ",
          err
        )
      }
    }

    //make sure we have both editors up-to-date (create AceEditorsStorageHelper tabs for every user solution file)

    if (!AceEditorsStorageHelper.getEditor(tutorViewEditorPanelViewEditorId)) {

      const userSolution = getState().tutorViewSiteState.userSolution
      const pLangs = getState().pLangsState.pLangs
      const plang = pLangs.find(p => p.id === userSolution.pLangId)

      if (!plang) {
        Logger.error('[tutorViewSiteActions] could not find plant for user solution')
      }

      const tabs = userSolution.solutionFiles.map<AceStoredTabLike>(p => {
        return {
          id: p.id,
          content: p.content,
          language: plang.internalName
        }
      })

      AceEditorsStorageHelper.addEditorWithTabs(tutorViewEditorPanelViewEditorId, tabs)
    }

    if (!AceEditorsStorageHelper.getEditor(tutorViewUserSolutionEditorPanelViewEditorId)) {

      const tutorSolution = getState().tutorViewSiteState.tutorEditorSolution
      const pLangs = getState().pLangsState.pLangs
      const plang = pLangs.find(p => p.id === tutorSolution.pLangId)

      if (!plang) {
        Logger.error('[tutorViewSiteActions] could not find plant for tutor solution')
      }

      const tabs = tutorSolution.solutionFiles.map<AceStoredTabLike>(p => {
        return {
          id: p.id,
          content: p.content,
          language: plang.internalName
        }
      })

      AceEditorsStorageHelper.addEditorWithTabs(tutorViewUserSolutionEditorPanelViewEditorId, tabs)
    }

    const selectedCodeTemplate = getState().tutorViewSiteState.exercise.codeTemplates.find(p => p.pLangId === getState().tutorViewSiteState.userSolution.pLangId)

    if (selectedCodeTemplate) {

      if (!AceEditorsStorageHelper.getEditor(tutorViewCodeTemplateEditorPanelViewEditorId)) {


        const pLangs = getState().pLangsState.pLangs
        const plang = pLangs.find(p => p.id === selectedCodeTemplate.pLangId)

        if (!plang) {
          Logger.error('[tutorViewSiteActions] could not find plant for code template')
        }

        const tabs = selectedCodeTemplate.templateFiles.map<AceStoredTabLike>(p => {
          return {
            id: p.id,
            content: p.content,
            language: plang.internalName
          }
        })

        AceEditorsStorageHelper.addEditorWithTabs(tutorViewCodeTemplateEditorPanelViewEditorId, tabs)
      }

      //select first file (if any)
      if (selectedCodeTemplate.templateFiles.length > 0
        && selectedCodeTemplate.templateFiles.filter(p => p.displayIndex >= 0).length > 0) {
        dispatch(setSelectedTabIdCodeTemplateEditorTutorView(selectedCodeTemplate.templateFiles.filter(p => p.displayIndex >= 0)[0].id))
      }

    }


    //try to select the first displayed tab (if any)
    if (getState().tutorViewSiteState.userSolution.solutionFiles.length > 0
      && getState().tutorViewSiteState.userSolution.solutionFiles.filter(p => p.uiIsDisplayed).length > 0) {
      dispatch(setSelectedTabIdUserSolutionEditorTutorView(
        getState().tutorViewSiteState.userSolution.solutionFiles.filter(p => p.uiIsDisplayed)[0].id))
    }

    //try to select the first displayed tab (if any) also for the tutor editor
    if (getState().tutorViewSiteState.tutorEditorSolution.solutionFiles.length > 0
      && getState().tutorViewSiteState.tutorEditorSolution.solutionFiles.filter(p => p.uiIsDisplayed).length > 0) {
      dispatch(setSelectedTabIdTutorEditor(
        getState().tutorViewSiteState.tutorEditorSolution.solutionFiles.filter(p => p.uiIsDisplayed)[0].id))
    }


    const userSolution = getState().tutorViewSiteState.userSolution

    if (userSolution.shouldLoadAfterSolution && userSolution.hasLoadAfterSolution) {
      await dispatch(getUserAfterSolutionAsTutorViewAsync(releaseId, userId, pLangId))


      if (!AceEditorsStorageHelper.getEditor(tutorViewUserAfterSolutionEditorPanelViewEditorId)) {

        const userAfterSolution = getState().tutorViewSiteState.userAfterSolution
        const pLangs = getState().pLangsState.pLangs
        const plang = pLangs.find(p => p.id === userAfterSolution.pLangId)

        if (!plang) {
          Logger.error('[tutorViewSiteActions] could not find plant for user after solution')
        }

        const tabs = userAfterSolution.solutionFiles.map<AceStoredTabLike>(p => {
          return {
            id: p.id,
            content: p.content,
            language: plang.internalName
          }
        })

        AceEditorsStorageHelper.addEditorWithTabs(tutorViewUserAfterSolutionEditorPanelViewEditorId, tabs)
      }

      if (getState().tutorViewSiteState.userAfterSolution.solutionFiles.length > 0
        && getState().tutorViewSiteState.userAfterSolution.solutionFiles.filter(p => p.uiIsDisplayed).length > 0) {
        dispatch(setSelectedTabIdAfterEditorTutorView(
          getState().tutorViewSiteState.userAfterSolution.solutionFiles.filter(p => p.uiIsDisplayed)[0].id))
      }

      await dispatch(getAllUserAfterTestResultsAsTutorAsnc(releaseId, userId, pLangId))

      if (!getState().tutorViewSiteState.exercise.exerciseRelease.isReleased
        && getState().tutorViewSiteState.exercise.exerciseRelease.hasAutomaticAssessmentFinished) {
        await dispatch(getAllAfterSubmitTestResultsAsTutor(releaseId, userId, pLangId))
      }
    }

    //create the assets
    //this is not working because we already set the exercise description and queried the wrong url...
    createAndRefreshAssetUrlsAndObjects(getState().tutorViewSiteState.exercise.exerciseDescription.assets)

    //then load the assessment...
    await dispatch(getSingleAssessmentTutorViewAsync(releaseId, userId, pLangId))

    //somehow force the preview panel to re render
    dispatch(setTutorViewDummy(false))

  }

}


export function resetUserSpecificDataAction(): RESET_userSpecificDataAction {
  return {
    type: ActionType.RESET_userSpecificData,
  }
}


export function refreshUserAssessmentData(releaseId: number, userId: number, pLangId: number,
                                          comesFromOwnExercisesSite: boolean
): MultiActions {
  return async (dispatch, getState) => {

    //we know here that the tutor site was already loaded & the submission site too

    dispatch(setComesFromOwnExercisesSite(comesFromOwnExercisesSite))

    //so we now only need to get the new assessment & user solution / data
    freeAssetUrlsAndObjects()
    setTutorViewDummy(true) //enable re rendering

    dispatch(resetUserSpecificDataAction())

    await dispatch(_loadTutorViewData(releaseId, userId, pLangId))

  }
}

export function setTutorViewDummy(dummy: boolean): SET_dummyAction {
  return {
    type: ActionType.SET_dummy,
    dummy
  }
}

export function setComesFromOwnExercisesSite(comesFromOwnExercisesSite: boolean): SET_comesFromOwnExercisesSiteAction {
  return {
    type: ActionType.SET_comesFromOwnExercisesSite,
    comesFromOwnExercisesSite
  }
}

export function setIsLoadingTutorViewSite(isLoading: boolean): SET_isLoadingAction {
  return {
    type: ActionType.SET_isLoading,
    isLoading
  }
}


export function resetTutorViewSite(): ResetAction {
  return {
    type: ActionType.RESET
  }
}
